/* v1.1, von Sebastian Voß, 28.04.2018
 * Dieses Programm gehört zu einem GPS-Quiz. Die Fragen erscheinen erst, wenn eine
 * bestimmte Position, eingestellt über GPS-Koordinaten, erreicht worden ist.
 * Danach kann per Multiple-Choice eine Antwort zur Frage ausgewählt werden.
 * Nach der letzten Frage dreht sich ein Servo-Motor, der als Schloss zum Gehäuse
 * angebracht ist. Somit kann das Gehäuse nach der letzten Frage geöffnet werden.
 * Im Gehäuse kann z.B. ein kleines Geschenk versteckt sein ("Reverse Geocache").
 * 
 */


#include <Wire.h>
#include "SSD1306Ascii.h"  
#include "SSD1306AsciiWire.h"
#include <TinyGPS++.h>
#include <SoftwareSerial.h>
#include <EEPROM.h>
#include <Servo.h>

/************* Festlegung der Pins und Erstellen der notwendigen Objekte etc **********/
// Pin configuration etc
static const int RXPin = 4, TXPin = 3;
static const byte tastereingang = A0;
static const byte taster_U_supply = 2;

static const uint32_t GPSBaud = 9600;
byte current_station = 0;
int eeprom_addr = 0;

// Servo object
Servo myservo;

// The TinyGPS++ object
TinyGPSPlus gps;
float lat_pos;                                   // speichert aktuellen Breitengrad
float lon_pos;                                  // speichert aktuellen Längengrad

// The serial connection to the GPS device
SoftwareSerial ss(RXPin, TXPin);

// Oled Display: Address and Object
#define I2C_ADDRESS 0x3C                      // Definition der I2C-Adresse des Displays
SSD1306AsciiWire oled;

#define no_t 100                              // Werte zur Kennzeichnung der gedrückten Taster
#define t1   101
#define t2   102
#define t3   103
#define t12  104
#define t13  105
#define t23  106
#define t123 107
uint8_t pressed;                              // Werte zur Speicherung des gedrückten Tasters


/** Im folgenden werden die Koordinaten, Fragen und Antwortmöglichkeiten festgelegt *************/
/*  Die Fragestrings haben eine maximale Länge von 19 Zeichen. Die Antwortstrings haben eine
 *  maximale Länge von 17 Zeichen.
 */

static const byte ANZ_FRAGEN = 3;                 // Anzahl der Fragen

static const float dest_lat_pos1 = 52.5256;     // Testkoordinaten
static const float dest_lon_pos1 = 7.3902;

String f1_z1 = "";                // Frage 1, Zeile 1
String f1_z2 = "";                // Frage 1, Zeile 2
String f1_a1 = "";                // Frage 1, Antwort 1
String f1_a2 = "";                // Frage 1, Antwort 2
String f1_a3 = "";                // Frage 1, Antwort 3
String f1_a4 = "";                // Frage 1, Antwort 4
byte f1_ra = 2;                   // Frage 1, richtige Antwort ist Antwort 2

static const float dest_lat_pos2 = 52.5227;     // Testkoordinaten
static const float dest_lon_pos2 = 7.4002;

String f2_z1 = ""; 
String f2_z2 = "";
String f2_a1 = "";
String f2_a2 = "";
String f2_a3 = "";
String f2_a4 = "";
byte f2_ra = 3;

static const float dest_lat_pos3 = 52.5268;     // Testkoordinaten
static const float dest_lon_pos3 = 7.4129;

String f3_z1 = ""; 
String f3_z2 = "";
String f3_a1 = "";
String f3_a2 = "";
String f3_a3 = "";
String f3_a4 = "";
byte f3_ra = 4;

void setup() 
{
  Serial.begin(115200);
  ss.begin(GPSBaud);
  Wire.begin();
  pinMode(taster_U_supply, OUTPUT);
  digitalWrite(taster_U_supply, HIGH);
  myservo.attach(13);
  myservo.write(50);                           // Startposition des Servo auf 50 Grad

  // *********** Start des Oled-Displays ************************************************************
  oled.begin(&Adafruit128x64, I2C_ADDRESS);
  //oled.set400kHz();                           // die Geschwindigkeit der I2C-Kommunikation lässt sich erhöhen, aber
                                                // ich weiß nicht, ob die Sensoren das aushalten
  oled.setFont(Adafruit5x7);                    // Festlegung der Schriftart
  oled.clear();                                 // Display wird geleert
  oled.setCursor(0,0);
  oled.println(F("Willkommen bei"));            // Willkommensnachricht
  oled.set2X();                                 // Nachricht doppelt so groß
  oled.println(F("Anonyms"));
  oled.println(F("   Quiz!"));
  oled.println();
  for(int i=0; i<3; i++) {
    oled.print(" .");
    /* Wenn die Taster links und rechts gleichzeitig gedrückt werden, öffnet sich ein Menü (das "BIOS"), 
    mit dem die Box manuell geöffnet und der EEPROM gelöscht werden kann, sodass man erneut spielen kann */
    pressed = no_t;
    int k=0;
    do
    {
      pressed=readtaster();
      delay(10);
      k++;
    } while (pressed != t13 && k<100);
    delay(10);
    if (pressed == t13) activate_BIOS();
  }
  oled.set1X();                                  // Nachricht wieder normal groß
  oled.clear();

  /*0: Position ermitteln */
  oled.println(F("Ermittle Position.."));
  oled.println(F("Kann etwas dauern"));

  // rufe solange die Position ab, bis gültige Werte erscheinen
  do{
    while (ss.available() > 0)
    {
      if (gps.encode(ss.read()))
      {
        oled.setCursor(0,2);
        displayGPSInfo();
      }
    }
  }
  while(!gps.location.isValid());
  // speichere Position
  lat_pos = gps.location.lat();
  lon_pos = gps.location.lng();
  oled.println(F("Geschafft!"));
  for(int i=0; i<3; i++)
  {
    oled.print(".");
    delay(1000);
  }

} // Ende des setups


void loop() {

  /* Ermittlung der aktuellen Station 
   * Wenn a noch nicht hochgesetzt wurde, wird das Programm zum ersten Mal abgespielt
   * und dann steht current_station richtigerweise auf Null. Es scheint so, dass
   * am Anfang (wenn noch ncihts von diesem Programm in den EEPROM geschrieben wurde) 
   * unter der EEPROM Adresse nicht immer Null steht, weshalb die Station Null
   * nicht aus dem EEPROM abgerufen werden soll.
   */
  byte a = EEPROM.read(eeprom_addr);
  if (a == 1 || a==2 || a== 3) current_station=a;

  switch (current_station)
  {
    case 0:
    {
      arrive_at_pos(/*nr*/ 0, /*dest_lat_pos*/ dest_lat_pos1, /*dest_lon_pos*/ dest_lon_pos1);
      
      quiz(/*quiz_nr*/ 0, /*question_row1*/ f1_z1, /*question_row2*/ f1_z2, 
           /*answer1*/ f1_a1, /*answer2*/ f1_a2, /*answer3*/ f1_a3, 
           /*answer4*/ f1_a4, /*right_answer_nr*/ f1_ra);
      break;
    }
    case 1:
    {
      arrive_at_pos(1, dest_lat_pos2, dest_lon_pos2);

      quiz(/*quiz_nr*/ 1, /*question_row1*/ f2_z1, /*question_row2*/ f2_z2, 
           /*answer1*/ f2_a1, /*answer2*/ f2_a2, /*answer3*/ f2_a3, 
           /*answer4*/ f2_a4, /*right_answer_nr*/ f2_ra);
            
      break;
    }
    case 2:
    {
      arrive_at_pos(2, dest_lat_pos3, dest_lon_pos3);        

      quiz(/*quiz_nr*/ 2, /*question_row1*/ f3_z1, /*question_row2*/ f3_z2, 
           /*answer1*/ f3_a1, /*answer2*/ f3_a2, /*answer3*/ f3_a3, 
           /*answer4*/ f3_a4, /*right_answer_nr*/ f3_ra);
      break;
    }
    case 3:
    {
      /* 6: Öffne die Box */
      myservo.write(110);
      delay(500);
      myservo.detach();
      oled.clear();
      oled.setCursor(0,1);
      oled.println("    Oeffne mich!   ");
      do{ } while (1);                            // Programmende 
      break;
    }
  }
  
} // Ende des loops

/**
 * Die GPS-Funktion zum Ankommen an einer bestimmten Zielposition nimmt die GPS-Koordinaten der Zielposition
 * sowie die zugehörige Nummer der Frage entgegen. Es wird die Zielposition und die aktuelle Position angezeigt
 * und ein Hinweis gegeben, in welcher Entfernung und Himmelsrichtung das Ziel liegt. Dies geschieht so lange,
 * bis man an der Zielposition angekommen ist.
 */

void arrive_at_pos(byte nr, float dest_lat_pos, float dest_lon_pos)
{
  float lat_dev, lon_dev;
  /* 1: Display: Aufforderung zu Koordinaten zu gehen und Aktualisierung der aktuellen Position */
  oled.clear();
  oled.println(F("Ziel fuer die ")); 
  oled.print(nr+1); oled.println(F(". Frage:"));
  oled.print(F("Lat: ")); oled.println(dest_lat_pos,6);
  oled.print(F("Lon: ")); oled.println(dest_lon_pos,6);
  oled.println(F("Deine Koordinaten:"));
  do 
  {
    oled.setCursor(0,5);
    oled.print(F("Lat:"));
    oled.println(lat_pos,5);
    oled.print(F("Lon:"));
    oled.println(lon_pos,5);
    
    
    oled.setCursor(0,7);
    // Bestimmung des Abstandes zum Ziel (Luftlinie)
    // 1: Überführung in ein lokales kartesisches Koordinatensystem (x,y)
    //    mit (0,0) = (dest_lon_pos, dest_lat_pos)
    float xc = (lon_pos - dest_lon_pos) * 3.141592 * 6371 * cos(lat_pos * 3.141592 / 180) / 180;
    float yc = (lat_pos - dest_lat_pos) * 3.141592 * 6371 / 180;

    // 2: Berechnung des Abstands in Metern (oben wurde mit dem Erdradius 6371 km gerechnet)
    float dist = sqrt(xc*xc + yc*yc);

    // 3: Ausgabe
    oled.print("Ziel ");
    if ( dist < 1) {
      dist *= 1000;
      oled.print(dist,0); oled.print("m");
    } else {
      oled.print(dist,0); oled.print("km");
    }
    
    oled.print(F(" in "));
    
    // Ausgabe der Zielrichtung
    lat_dev = lat_pos - dest_lat_pos;                       // deviation in latitude
    lon_dev = lon_pos - dest_lon_pos;                       // deviation in longitude
    if ( lat_dev > 0.0005 && lon_dev > 0.0005 )
    {
      oled.print(F("SW"));
    }
    else if ( lat_dev < -0.0005 && lon_dev > 0.0005 )
    {
      oled.print(F("NW"));
    }
    else if ( lat_dev > 0.0005 && lon_dev < -0.0005 )
    {
      oled.print(F("SO"));
    }
    else if ( lat_dev < -0.0005 && lon_dev < -0.0005 )
    {
      oled.print(F("NO"));
    }
    else if ( lat_dev > 0.0005 && abs(lon_dev) < 0.0005 )
    {
      oled.print(F("S"));
    }
    else if ( lat_dev < -0.0005 && abs(lon_dev) < 0.0005 )
    {
      oled.print(F("N"));
    }
    else if ( abs(lat_dev) < 0.0005 && lon_dev > 0.0005 )
    {
      oled.print(F("W"));
    }
    else if ( abs(lat_dev) < 0.0005 && lon_dev < -0.0005 )
    {
      oled.print(F("O"));
    }

    delay(1000);
    
    // Neubestimmung der Position
    while (ss.available() > 0)
    {
      if (gps.encode(ss.read()))
      {
        if ( gps.location.isValid())
        {
          lat_pos= gps.location.lat();
          lon_pos= gps.location.lng();
        }
      }
    } 
    
  } while (lat_dev > 0.0005 || lat_dev < -0.0005 || lon_dev > 0.0005 || lon_dev < -0.0005);
}

/**
 * Die Quiz-Funktion zeigt das Quiz an und verarbeitet die Antwort. Als Argumente werden die Frage,
 * die Antwortmöglichkeiten sowie die Nummer der richtigen Antwort entgegengenommen.
 * Die Frage wird solange wiederholt, bis die richtige Antwort gegeben wurde.
 */
 
void quiz(byte quiz_nr, String question_row1, String question_row2, 
          String answer1, String answer2, String answer3, String answer4, 
          byte right_answer_nr)
{
  /* 2: Die Frage */
  oled.clear();
  oled.println(F("Ziel erreicht!"));
  oled.println(F("Nun die Frage"));
  oled.print(F("             "));
  for ( int i =0; i<3; i++) 
  {
    oled.print(".");
    delay(1000);
  }
  
  do {
    oled.clear();
    oled.println(question_row1);
    oled.println(question_row2);
    oled.print(F("  ")); oled.println(answer1);
    oled.print(F("  ")); oled.println(answer2);
    oled.print(F("  ")); oled.println(answer3);
    oled.print(F("  ")); oled.println(answer4);
    oled.println();
    oled.println();
    printuparrow();
    oled.print(F("        OK        "));
    printdownarrow();
    
    pressed = no_t;
    uint8_t index=0;                              // Index für die Position des Pfeils
    oled.setCursor(0,index+2);
    oled.print(">");
    while (pressed != t2) 
    {                                             // solange nicht OK gedrückt wurde
      //pressed=readtaster();                       // wichtig für den zweiten Durchlauf, sonst wird pressed nie erneuert
      do 
      {                                           // warte auf Eingabe
        pressed=readtaster();
        delay(10);
      } while ( pressed==no_t);
      delay(10);
      if (pressed == t2) break;                   // wenn t2, also OK, gedrückt wurde, springe sofort raus
      oled.setCursor(0,index+2);                  // Sonst wird die Pfeilposition angepasst; zunächst lösche alten Pfeil
      oled.print(" ");
      switch (pressed) {                          // ansonsten unterscheide ob es hoch oder runter geht
        case t3:                                  // wenn t1 gedrückt wurde, wird der Index um 1 verringert
          if(index >0) {                          // allerdings nur, wenn der Index nicht schon Null ist
            index--;
          } else {                                // sonst bleibt er null
            index=0;
          }
          break;
        case t1:                                  // analog für t3
          if(index <3) {
            index++;
          } else {
            index=3;
          }
          break;
      }
      oled.setCursor(0,index+2);                  // setze neuen Pfeil
      oled.print(">");
      delay(500);
    }
    oled.clear();
    
    /* 3: Verarbeitung der Antwort */
    if ( index == right_answer_nr -1)           // richtige Antwort (da ich mit 1 statt 0 anfange zu zählen, muss 1 abgezogen werden)
    {
      oled.println(F("Richtig!"));
      current_station++;                        // Station geschafft, also ist die nächste Station jetzt aktuell
      EEPROM.write(eeprom_addr,current_station);  // Speichern im EEPROM, damit der Arduino ausgestellt werden kann
    }
    else                                        // falsche Antwort
    {
      oled.println(F("Leider falsch!"));
      oled.println(F("Gleich noch mal"));
    }
    oled.print(F("          "));                // Übergangsanimation
    for( int i=0; i<3; i++) 
    {
      oled.print(".");
      delay(1000);
    }
    
    oled.clear();
  } while ( current_station == quiz_nr);
  // solange nicht die richtige Antwort gegeben wurde, wird die aktuelle Station nicht hochgesetzt
}


/** 
 * Ausgabe der eingelesenen GPS-Informationen
 */
 
void displayGPSInfo()
{
  oled.println(F("Position: ")); 
  if (gps.location.isValid())
  {
    oled.print(F("Lat: "));
    oled.println(gps.location.lat(), 6);
    oled.print(F("Lon: "));
    oled.println(gps.location.lng(), 6);
  }
  else
  { 
    oled.print(F("Lat: "));
    oled.println(F("INVALID"));
    oled.print(F("Lon: "));
    oled.println(F("INVALID"));
  }

}

/**
 * Funktion zum Auslesen des gedrückten Tasters
 * Konfiguration bzw. Werte im analogen Eingang:
 * Kein Taster:                     666
 * Taster 1 (parallel zu 330 Ohm):  620
 * Taster 2 (parallel zu 530 Ohm):  585
 * mitte und rechts gedrueckt: analogRead -> 508
 * Taster 3 (parallel zu 1000 Ohm): 475
 * links und rechts gedrueckt: analogRead -> 358
 * links und mitte gedrueckt: analogRead -> 251
 * alle drei Taster gedrueckt: analogRead -> 0
 */

uint8_t readtaster(void) {
  uint16_t key_in = analogRead(tastereingang);
  if (key_in > 640) return no_t;
  if (key_in > 602) return t1;
  if (key_in > 530) return t2;
  if (key_in > 490) return t12;
  if (key_in > 430) return t3;
  if (key_in > 300) return t13;
  if (key_in > 200) return t23;
  return t123;
}


/**
 * Pfeilsymbol, hier nach unten zeigend 
 */
 
void printdownarrow(void) {
  Wire.beginTransmission(0x3C);  // Display Adresse
  Wire.write(0x40);              // Data Mode - es folgen Daten
  Wire.write(0x0C);              // (0x)0C = 0000 1100
  Wire.write(0x18);              // (0x)18 = 0001 1000
  Wire.write(0x30);              // (0x)30 = 0011 0000
  Wire.write(0x60);              // (0x)60 = 0110 0000
  Wire.write(0x60);              // (0x)60 = 0110 0000
  Wire.write(0x30);              // (0x)30 = 0011 0000
  Wire.write(0x18);              // (0x)18 = 0001 1000
  Wire.write(0x0C);              // (0x)0C = 0000 1100
  Wire.endTransmission();        // Beenden der Übertragung
}

/**
 * Pfeilsymbol, hier nach oben zeigend 
 */
 
void printuparrow(void) {
  Wire.beginTransmission(0x3C);  // Display Adresse
  Wire.write(0x40);              // Data Mode - es folgen Daten
  Wire.write(0x30);              // (0x)30 = 0011 0000
  Wire.write(0x18);              // (0x)18 = 0001 1000
  Wire.write(0x0C);              // (0x)0C = 0000 1100
  Wire.write(0x06);              // (0x)06 = 0000 0110
  Wire.write(0x06);              // (0x)06 = 0000 0110
  Wire.write(0x0C);              // (0x)0C = 0000 1100
  Wire.write(0x18);              // (0x)18 = 0001 1000
  Wire.write(0x30);              // (0x)30 = 0011 0000
  Wire.endTransmission();        // Beenden der Übertragung
}

/**
 * Das "BIOS"-Menü lässt sich öffnen, wenn man während des Willkommensbildschirms den linken und rechten Taster
 * drückt. Darin lassen sich "administrative Funktionen" erreichen, falls das GPS-Quiz nicht richtig funktionieren sollte.
 * Es lässt sich die Box manuell öffnen, der EEPROM (in dem die aktuelle Station bzw. Frage gespeichert wird) zurücksetzen
 * und die aktuelle Station bzw. Frage manuell anpassen.
 */
 
void activate_BIOS (void)
{
  oled.clear();
  oled.set1X();                                  // Nachricht wieder normal groß
  oled.println();
  oled.println();
  oled.print(F("  ")); oled.println("Box oeffnen");
  oled.print(F("  ")); oled.println("EEPROM leeren");
  oled.print(F("  ")); oled.println("Setze Station");
  oled.print(F("  ")); oled.println("Abbruch");
  oled.println();
  oled.println();
  printuparrow();
  oled.print(F("        OK        "));
  printdownarrow();
  
  pressed = no_t;
  uint8_t index=0;                              // Index für die Position des Pfeils
  oled.setCursor(0,index+2);
  oled.print(">");
  while (pressed != t2) 
  {                                             // solange nicht OK gedrückt wurde
    //pressed=readtaster();                       // wichtig für den zweiten Durchlauf, sonst wird pressed nie erneuert
    do 
    {                                           // warte auf Eingabe
      pressed=readtaster();
      delay(10);
    } while ( pressed==no_t);
    delay(10);
    if (pressed == t2) break;                   // wenn t2, also OK, gedrückt wurde, springe sofort raus
    oled.setCursor(0,index+2);                  // Sonst wird die Pfeilposition angepasst; zunächst lösche alten Pfeil
    oled.print(" ");
    switch (pressed) {                          // ansonsten unterscheide ob es hoch oder runter geht
      case t3:                                  // wenn t1 gedrückt wurde, wird der Index um 1 verringert
        if(index >0) {                          // allerdings nur, wenn der Index nicht schon Null ist
          index--;
        } else {                                // sonst bleibt er null
          index=0;
        }
        break;
      case t1:                                  // analog für t3
        if(index <3) {
          index++;
        } else {
          index=3;
        }
        break;
    }
    oled.setCursor(0,index+2);                  // setze neuen Pfeil
    oled.print(">");
    delay(500);
  }
  switch (index)
  {
    case 0:
      open_box();
      break;
    case 1:
      clear_eeprom();
      break;
    case 2:
      set_station();
      break;
    default: break;
  }
  oled.clear();
}

/**
 * "BIOS"-Funktion. Öffnet die Box nach manueller Aufforderung.
 */
 
void open_box()
{
  myservo.write(110);
  delay(500);
  myservo.detach();
  do {} while(1);
}

/**
 * "BIOS"-Funktion. Überschreibt den EEPROM mit lauter Nullen, sodass insbesondere
 * die aktuelle Station/Frage wieder auf null (auf den Anfang) gesetzt wird.
 */
 
void clear_eeprom()
{
  for (int i = 0 ; i < EEPROM.length() ; i++) {
    EEPROM.write(i, 0);
  }
}

/**
 * "BIOS"-Funktion. Erlaubt das manuelle Einstellen der aktuellen Station/Frage.
 */
 
void set_station()
{
  pressed = no_t;
  do {
    // Bildschirmausgabe
    oled.clear();
    oled.setCursor(0,2);
    oled.println(F("Aktuelle Station:"));
    oled.println("   " + String(current_station));
    oled.setCursor(0,7);
    oled.println("--      OK      ++");
    delay(250);
    
    // warte auf Eingabe
    do {
      pressed = readtaster();
      delay(10);
    } while ( pressed == no_t);

    // Verarbeitung der Eingabe
    switch(pressed) {
      case t3:
        if(current_station >0) { 
          current_station--;
          EEPROM.write(eeprom_addr,current_station);
        }
        break;
      case t2:
        break;
      case t1:
        if(current_station < ANZ_FRAGEN -1) {
          current_station++;
          EEPROM.write(eeprom_addr, current_station);
        }
        break;
      default: break;
    }
  } while( pressed != t2);
}














